# NOTAZIONE PREFISSO
# la notazione infissa (l'operatore e' tra gli operandi) per scrivere 
# espressioni aritmetiche e' ambigua. Per elminare l'ambiguita' bisogna 
# inserire parentesi o introdurre priorita' tra gli operandi.
# 1 + 2 * 3 + 5 = 12
# la notazione prefissa (l'operatore recede i suoi operandi) + + 1 * 2 3 5
# non produce ambiguita'  
#  1 + (2 * 3) + 4  ----->  + + 1 * 2 3 4
#  (1+ 2) * (3 + 5 )----->  * + 1 2 + 3 4 

#################################################

# ESERCIZIO: da espressione prefissa ottieni espressione infissa

#IDEA: costruisco l’ALBERO DELL'ESPRESSIONE (i nodi foglia sono i numeri e 
#quelli interni le operazioni tutte binarie e quindi ogni nodo interno 
#ha esattamente due figli) e poi lo visito e ne stampo i nodi. 
#Per ottenere la notazione infissa basta stampare ricorsivamente 
#prima i nodi del sottoalbero sinistro, poi il nodo radice e poi i nodi 
# del sottoalbero destro (visita inorder dell’albero binario).

def da_prefissa_a_infissa(s):
    # scomponiamo l'espressione nei suoi componenti 
    lista=s.split()
    # creiamo l'albero con l'espressione
    rad=crea_albero(lista) 
    return visita_inorder(rad)



class Nodo:
    def __init__(self, x):
        self.nome = x
        self.fs =None
        self.fd= None


def crea_albero(lista):
    ''' data la lista ordinata degli elementi della formula 
    prefissa crea l'albero dell'espressione'''
    nodo=Nodo(lista[0])
    del lista[0]
    if not nodo.nome.isdigit():
        nodo.fs=crea_albero(lista)
        nodo.fd=crea_albero(lista)
    return nodo

        
def visita_inorder(nodo):
    '''restituisce la stringa con i nodi visitati in modo inorder'''
    if nodo.nome.isdigit(): return ' '+ nodo.nome +' '
    return visita_inorder(nodo.fs)+nodo.nome+visita_inorder(nodo.fd)

# >>> da_prefissa_a_infissa('+ 30 45')
# ' 30 + 45 '

# >>> da_prefissa_a_infissa('- + + 1 2 3 4')
# ' 1 + 2 + 3 - 4 '


# >>> da_prefissa_a_infissa('+ 30 45')
# ' 30 + 45 '


# >>> da_prefissa_a_infissa('+ + 1 * 2 3 4')
# ' 1 + 2 * 3 + 4 '

# >>> da_prefissa_a_infissa('* + 1 2 + 3 4')
# ' 1 + 2 * 3 + 4 ' Scorretta
            
#>>> da_prefissa_a_infissa('* + 3 5 - 7 2')
# ' 3 + 5 * 7 - 2 '  Scorretta

############################################################
 
#MEGLIO PARENTESIZZARE L'ESPRESSIONE INORDER PER EVITARE AMBIGUITA' 
 
def visita_inorder1(nodo):
    '''restituisce la stringa con i nodi visitati in modo inorder'''
    if nodo.nome.isdigit(): return ' '+ nodo.nome +' '
    return '('+ visita_inorder1(nodo.fs) + nodo.nome + \
    visita_inorder1(nodo.fd)+ ')'

def da_prefissa_a_infissa1(s):
    # scomponiamo l'espressione nei suoi componenti 
    lista=s.split()
    # creiamo l'albero con l'espressione
    rad=crea_albero(lista) 
    return visita_inorder1(rad)

# >>> da_prefissa_a_infissa1('+ 30 45')
# '( 30 + 45 )'

# >>> da_prefissa_a_infissa1('- + + 1 2 3 4')
# '((( 1 + 2 )+ 3 )- 4 )'


# >>> da_prefissa_a_infissa1('+ + 1 * 2 3 4')
# '(( 1 +( 2 * 3 ))+ 4 )'

# >>> da_prefissa_a_infissa1('* + 1 2 + 3 4')
# '(( 1 + 2 )*( 3 + 4 ))'
            
#>>> da_prefissa_a_infissa1('* + 3 5 - 7 2')
# '(( 3 + 5 )*( 7 - 2 ))'
           
###############################################################################           
#una volta costruito l'albero dell'espressione e'anche facile 
# valutare il valore dell'espressione sempre grazia a una 
#visita inorder

def valuta(rad):
    #data la radice di un'albero dell'espressione 
    #ne restituisce il valore
    if rad.nome.isdigit(): return int(rad.nome)
    a=valuta(rad.fs)
    b=valuta(rad.fd)
    if rad.nome=='+': v= a+b
    elif rad.nome=='-': v= a-b
    elif rad.nome=='*': v= a*b
    return v 

# >>> valuta(crea_albero('+ + 1 * 2 3 4'.split()))
# 11
    
# >>> valuta(crea_albero('* + 1 2 + 3 4'.split()))
# 21
             
###############################################################################
    
def da_prefissa_a_postfissa(s):
    # scomponiamo l'espressione nei suoi componenti 
    lista=s.split()
    # creiamo l'albero con l'espressione
    rad=crea_albero(lista) 
    return visita_postorder(rad)

def visita_postorder(nodo):
    '''restituisce la stringa con i nodi visitati in modo inorder'''
    if nodo.nome.isdigit(): return ' '+ nodo.nome +' '
    return visita_postorder(nodo.fs)+ visita_postorder(nodo.fd)+nodo.nome

# >>> da_prefissa_a_postfissa('* + 1 2 + 3 4')
# >>>  ' 1  2 + 3  4 +*'